# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple
from datetime import datetime
from getpass import getuser
from os import environ

import pytest

from v3iofs import V3ioFS

host = environ.get('V3IO_API')
access_key = environ.get('V3IO_ACCESS_KEY')
container = 'bigdata'  # TODO: configure


Obj = namedtuple('Obj', 'path data')


@pytest.fixture
def fs():
    fs = V3ioFS(v3io_api=host, v3io_access_key=access_key)
    yield fs
    fs._client.close()


@pytest.fixture
def tmp_obj():
    user, ts = getuser(), datetime.now().isoformat()
    client = V3ioFS(v3io_api=host, v3io_access_key=access_key)._client

    path = f'{user}-test-{ts}'
    body = f'test data for {user} at {ts}'
    client.put_object(container, path, body=body.encode())

    yield Obj(f'/{container}/{path}', body)

    client.delete_object(container, path)



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ssl
from urllib.request import Request, urlopen

from fsspec.spec import AbstractBufferedFile

# Ignore certs
no_validate_ctx = ssl.create_default_context()
no_validate_ctx.check_hostname = False
no_validate_ctx.verify_mode = ssl.CERT_NONE

# v3io-py don't support range requests so we're using urllib
# (https://github.com/v3io/v3io-py/issues/16)


class V3ioFile(AbstractBufferedFile):
    def _fetch_range(self, start, end):
        req = Request(
            url=f'{self.fs._v3io_api}/{self.path}',
            headers={
                'Range': f'bytes={start}-{end}',
                'X-v3io-session-key': self.fs._v3io_access_key,
            },
        )
        with urlopen(req, context=no_validate_ctx) as fp:
            return fp.read()

    def _upload_chunk(self, final=False):
        """ Write one part of a multi-block file upload

        Parameters
        ==========
        final: bool
            This is the last block, so should complete file, if
            self.autocommit is True.
        """
        data = self.buffer.getvalue()
        if not data:
            return
        req = Request(
            method='POST',
            data=data,
            url=f'{self.fs._v3io_api}/{self.path}',
            headers={
                'Range': '-1',
                'Content-Length': len(data),
                'X-v3io-session-key': self.fs._v3io_access_key,
            },
        )
        urlopen(req, context=no_validate_ctx)



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""v3iofs - An fsspec driver for v3io"""

__all__ = [
    '__version__',
    'V3ioFS',
    'V3ioFile',
]

from fsspec.registry import known_implementations

from .file import V3ioFile  # noqa: F401
from .fs import V3ioFS  # noqa: F401

__version__ = '0.1.0a1'

# Register v3io protocol
# (https://github.com/intake/filesystem_spec/issues/293)
if V3ioFS.protocol not in known_implementations:
    known_implementations[V3ioFS.protocol] = {
        'class': 'v3iofs.V3ioFS',
    }

del known_implementations  # clear the module namespace



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import dask.bag as db

from conftest import access_key, host

data = 'In god we trust; all others must bring data.'


def test_dask(tmp_obj):
    uri = f'v3io://{host}/{tmp_obj.path}'
    storage_options = {
        'V3IO_ACCESS_KEY': access_key
    }
    file = db.read_text(uri, storage_options=storage_options)
    data, = file.compute()
    assert tmp_obj.data == data, 'bad data'



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime, timezone
from os.path import dirname

import pytest

from v3iofs import V3ioFS
from v3iofs.fs import parse_time, split_auth
from v3iofs.path import split_container

container = 'bigdata'  # FIXME


def test_ls(fs: V3ioFS):
    path = f'/{container}/miki/'
    out = fs.ls(path)
    assert len(out) > 0, 'nothing found'
    assert all(isinstance(p, dict) for p in out), 'not dict'

    out = fs.ls(path, detail=False)
    assert len(out) > 0, 'nothing found'
    assert all(isinstance(p, str) for p in out), 'not string'


def test_rm(fs: V3ioFS, tmp_obj):
    path = tmp_obj.path
    fs.rm(path)
    out = fs.ls(dirname(path), detail=False)
    assert path not in out, 'not deleted'


def test_touch(fs: V3ioFS, tmp_obj):
    path = tmp_obj.path
    fs.touch(path)
    container, path = split_container(path)
    resp = fs._client.get_object(container, path)
    assert resp.body == b'', 'not truncated'


now = datetime(2020, 1, 2, 3, 4, 5, 6789, tzinfo=timezone.utc)
parse_time_cases = [
    # value, expected, raises
    ('', None, True),
    (now.strftime('%Y-%m-%d'), None, True),
    (now.strftime('%Y-%m-%dT%H:%M:%S.%f%z'), now.timestamp(), False),
    (now.strftime('%Y-%m-%dT%H:%M:%S.%fZ'), now.timestamp(), False),
]


@pytest.mark.parametrize('value, expected, raises', parse_time_cases)
def test_parse_time(value, expected, raises):
    if raises:
        with pytest.raises(ValueError):
            parse_time(value)
        return

    out = parse_time(value)
    assert expected == out


split_auth_cases = [
    (
        'v3io://api_key:s3cr3t@domain.company.com',
        ('v3io://domain.company.com', 's3cr3t'),
        False,
    ),
    (
        'v3io://domain.company.com',
        ('v3io://domain.company.com', ''),
        False,
    ),
    (
        'v3io://s3cr3t@domain.company.com',
        ('v3io://domain.company.com', ''),
        True,
    ),
]


@pytest.mark.parametrize('url, expected, raises', split_auth_cases)
def test_split_auth(url, expected, raises):
    if raises:
        with pytest.raises(ValueError):
            split_auth(url)
        return
    out = split_auth(url)
    assert expected == out



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime, timezone
from http import HTTPStatus
from os import environ
from urllib.parse import urlparse

from fsspec.spec import AbstractFileSystem
from v3io.dataplane import Client

from .path import split_container, unslash


class V3ioFS(AbstractFileSystem):
    """File system driver to v3io

    Parameters
    ----------
    v3io_api: str
        API host name (or V3IO_API environment)
    v3io_access_key: str
        v3io access key (or V3IO_ACCESS_KEY from environment)
    **kw:
        Passed to fsspec.AbstractFileSystem
    """

    protocol = 'v3io'

    def __init__(self, v3io_api=None, v3io_access_key=None, **kw):
        # TODO: Support storage options for creds (in kw)
        super().__init__(**kw)
        self._v3io_api, self._v3io_access_key = \
            self._parse_api(v3io_api, v3io_access_key)
        self._client = Client(
            endpoint=self._v3io_api,
            access_key=self._v3io_access_key,
            transport_kind='requests',
        )

    def ls(self, path, detail=True, **kwargs):
        """Lists files & directories under path"""
        container, path = split_container(path)

        if not container:
            return self._list_containers(detail)

        resp = self._client.get_container_contents(
            container=container,
            path=path,
            get_all_attributes=True,
            raise_for_status=[HTTPStatus.OK],
        )

        prefixes = resp.output.common_prefixes  # directories
        fn = prefix_info if detail else prefix_path
        dirs = [fn(container, p) for p in prefixes]

        objects = resp.output.contents  # files
        fn = object_info if detail else object_path
        files = [fn(container, o) for o in objects]

        return dirs + files

    def _parse_api(self, v3io_api, v3io_access_key):
        v3io_api = v3io_api or environ.get('V3IO_API')
        url, auth = split_auth(v3io_api)
        if auth:
            return url, auth

        v3io_access_key = v3io_access_key or environ.get('V3IO_ACCESS_KEY')
        return v3io_api, v3io_access_key

    def _list_containers(self, detail):
        resp = self._client.get_containers(
            raise_for_status=[HTTPStatus.OK],
        )
        fn = container_info if detail else container_path
        return [fn(c) for c in resp.output.containers]

    def copy(self, path1, path2, **kwargs):
        ...  # FIXME

    def _rm(self, path):
        container, path = split_container(path)
        if not container:
            raise ValueError(f'bad path: {path:r}')

        self._client.delete_object(
            container=container,
            path=path,
            raise_for_status=[HTTPStatus.NO_CONTENT],
        )

    def touch(self, path, truncate=True, **kwargs):
        if not truncate:  # TODO
            raise ValueError('only truncate touch supported')
        container, path = split_container(path)
        self._client.put_object(
            container, path,
            raise_for_status=[HTTPStatus.OK],
        )


def container_path(container):
    return f'/{container.name}'


def container_info(container):
    return {
        'name': container.name,
        'size': None,
        'created': parse_time(container.creation_date),
    }


def prefix_path(container_name, prefix):
    if not isinstance(prefix, str):
        prefix = prefix.prefix
    # prefix already have a leading /
    return unslash(f'/{container_name}{prefix}')


def prefix_info(container_name, prefix):
    return info_of(container_name, prefix, 'prefix')


def object_path(container_name, object):
    # object.key already have a leading /
    return f'/{container_name}{object.key}'


def object_info(container_name, object):
    return info_of(container_name, object, 'key')


def info_of(container_name, obj, name_key):
    return {
        'name': prefix_path(container_name, getattr(obj, name_key)),
        'size': getattr(obj, 'size', None),
        'created': parse_time(obj.creating_time),
        'mtime': parse_time(obj.last_modified),
        'atime': parse_time(obj.access_time),
        'mode': int(obj.mode[1:], base=8),  # '040755'
        'gid': int(obj.gid, 16),
        'uid': int(obj.uid, 16),
    }


def parse_time(creation_date):
    # '2020-03-26T09:42:57.504000+00:00'
    # '2020-03-26T09:42:57.71Z'
    i = creation_date.rfind('+')  # If not found will be -1, good for Z
    dt = datetime.strptime(creation_date[:i], '%Y-%m-%dT%H:%M:%S.%f')
    dt = dt.replace(tzinfo=timezone.utc)
    return dt.timestamp()


def split_auth(url):
    """
    >>> split_auth('v3io://api_key:s3cr3t@domain.company.com')
    ('v3io://domain.company.com', 's3cr3t')
    >>> split_auth('v3io://domain.company.com')
    ('v3io://domain.company.com', '')
    """
    u = urlparse(url)
    if '@' not in u.netloc:
        return (url, '')

    auth, netloc = u.netloc.split('@', 1)
    if ':' not in auth:
        raise ValueError('missing : in auth')
    _, key = auth.split(':', 1)
    u = u._replace(netloc=netloc)
    return (u.geturl(), key)



from subprocess import run
from sys import executable
from tempfile import mkdtemp

import v3iofs


def test_install():
    out = run([executable, 'setup.py', 'sdist'])
    assert out.returncode == 0, 'sdist'

    tmp_dir = mkdtemp(prefix='v3iofs-test')
    out = run([executable, '-m', 'venv', tmp_dir])
    assert out.returncode == 0, 'create venv'

    dist = f'./dist/v3iofs-{v3iofs.__version__}.tar.gz'
    py = f'{tmp_dir}/bin/python'

    out = run([py, '-m', 'pip', 'install', dist])
    assert out.returncode == 0, 'install'

    out = run([py, '-c', 'import v3iofs'])
    assert out.returncode == 0, 'import'



import fsspec
from v3iofs import V3ioFS


def test_register():
    cls = fsspec.get_filesystem_class(V3ioFS.protocol)
    assert cls is V3ioFS, 'not registered'

    options = {
        'v3io_api': 'a.b.com',
        'v3io_access_key': 's3cr3t',
    }
    fs = fsspec.filesystem('v3io', **options)
    assert isinstance(fs, V3ioFS), f'bad object class - {fs.__class__}'
